@@ -12,6 +12,7 @@
 import re
 
 import netaddr
+from oslo_utils import uuidutils
 
 from ovsdbapp.backend.ovs_idl import command as cmd
 from ovsdbapp.backend.ovs_idl import idlutils
@@ -80,26 +81,29 @@ class LsGetCommand(cmd.BaseGetRowCommand):
     table = 'Logical_Switch'
 
 
-class AclAddCommand(cmd.AddCommand):
+class _AclAddHelper(cmd.AddCommand):
     table_name = 'ACL'
 
-    def __init__(self, api, switch, direction, priority, match, action,
-                 log=False, may_exist=False, **external_ids):
+    def __init__(self, api, entity, direction, priority, match, action,
+                 log=False, may_exist=False, severity=None, name=None,
+                 **external_ids):
         if direction not in ('from-lport', 'to-lport'):
             raise TypeError("direction must be either from-lport or to-lport")
         if not 0 <= priority <= const.ACL_PRIORITY_MAX:
-            raise ValueError("priority must be beween 0 and %s, inclusive" % (
+            raise ValueError("priority must be between 0 and %s, inclusive" % (
                              const.ACL_PRIORITY_MAX))
         if action not in ('allow', 'allow-related', 'drop', 'reject'):
             raise TypeError("action must be allow/allow-related/drop/reject")
-        super(AclAddCommand, self).__init__(api)
-        self.switch = switch
+        super(_AclAddHelper, self).__init__(api)
+        self.entity = entity
         self.direction = direction
         self.priority = priority
         self.match = match
         self.action = action
         self.log = log
         self.may_exist = may_exist
+        self.severity = severity
+        self.name = name
         self.external_ids = external_ids
 
     def acl_match(self, row):
@@ -108,8 +112,8 @@ class AclAddCommand(cmd.AddCommand):
                 self.match == row.match)
 
     def run_idl(self, txn):
-        ls = self.api.lookup('Logical_Switch', self.switch)
-        acls = [acl for acl in ls.acls if self.acl_match(acl)]
+        entity = self.api.lookup(self.lookup_table, self.entity)
+        acls = [acl for acl in entity.acls if self.acl_match(acl)]
         if acls:
             if self.may_exist:
                 self.result = rowview.RowView(acls[0])
@@ -122,45 +126,183 @@ class AclAddCommand(cmd.AddCommand):
         acl.match = self.match
         acl.action = self.action
         acl.log = self.log
-        ls.addvalue('acls', acl)
+        acl.severity = self.severity
+        acl.name = self.name
+        entity.addvalue('acls', acl)
         for col, value in self.external_ids.items():
             acl.setkey('external_ids', col, value)
         self.result = acl.uuid
 
 
-class AclDelCommand(cmd.BaseCommand):
+class AclAddCommand(_AclAddHelper):
+    lookup_table = 'Logical_Switch'
+
+    def __init__(self, api, switch, direction, priority, match, action,
+                 log=False, may_exist=False, severity=None, name=None,
+                 **external_ids):
+        # NOTE: we're overriding the constructor here to not break any
+        # existing callers before we introduced Port Groups.
+        super(AclAddCommand, self).__init__(api, switch, direction, priority,
+                                            match, action, log, may_exist,
+                                            severity, name, **external_ids)
+
+
+class PgAclAddCommand(_AclAddHelper):
+    lookup_table = 'Port_Group'
+
+
+class _AclDelHelper(cmd.BaseCommand):
+    def __init__(self, api, entity, direction=None,
+                 priority=None, match=None):
+        if (priority is None) != (match is None):
+            raise TypeError("Must specify priority and match together")
+        if priority is not None and not direction:
+            raise TypeError("Cannot specify priority/match without direction")
+        super(_AclDelHelper, self).__init__(api)
+        self.entity = entity
+        self.conditions = []
+        if direction:
+            self.conditions.append(('direction', '=', direction))
+            # priority can be 0
+            if match:  # and therefore priority due to the above check
+                self.conditions += [('priority', '=', priority),
+                                    ('match', '=', match)]
+
+    def run_idl(self, txn):
+        entity = self.api.lookup(self.lookup_table, self.entity)
+        for acl in [a for a in entity.acls
+                    if idlutils.row_match(a, self.conditions)]:
+            entity.delvalue('acls', acl)
+            acl.delete()
+
+
+class AclDelCommand(_AclDelHelper):
+    lookup_table = 'Logical_Switch'
+
+    def __init__(self, api, switch, direction=None,
+                 priority=None, match=None):
+        # NOTE: we're overriding the constructor here to not break any
+        # existing callers before we introduced Port Groups.
+        super(AclDelCommand, self).__init__(api, switch, direction, priority,
+                                            match)
+
+
+class PgAclDelCommand(_AclDelHelper):
+    lookup_table = 'Port_Group'
+
+
+class _AclListHelper(cmd.BaseCommand):
+    def __init__(self, api, entity):
+        super(_AclListHelper, self).__init__(api)
+        self.entity = entity
+
+    def run_idl(self, txn):
+        entity = self.api.lookup(self.lookup_table, self.entity)
+        self.result = [rowview.RowView(acl) for acl in entity.acls]
+
+
+class AclListCommand(_AclListHelper):
+    lookup_table = 'Logical_Switch'
+
+
+class PgAclListCommand(_AclListHelper):
+    lookup_table = 'Port_Group'
+
+
+class QoSAddCommand(cmd.AddCommand):
+    table_name = 'QoS'
+
+    def __init__(self, api, switch, direction, priority, match, rate=None,
+                 burst=None, dscp=None, may_exist=False, **columns):
+        if direction not in ('from-lport', 'to-lport'):
+            raise TypeError("direction must be either from-lport or to-lport")
+        if not 0 <= priority <= const.ACL_PRIORITY_MAX:
+            raise ValueError("priority must be between 0 and %s, inclusive" %
+                             const.ACL_PRIORITY_MAX)
+        if rate is not None and not 1 <= rate <= const.QOS_BANDWIDTH_MAX:
+            raise ValueError("rate(%s) must be between 1 and %s, inclusive" %
+                             rate, const.QOS_BANDWIDTH_MAX)
+        if burst is not None and not 1 <= burst <= const.QOS_BANDWIDTH_MAX:
+            raise ValueError("burst(%s) must be between 1 and %s, "
+                             "inclusive" % burst, const.QOS_BANDWIDTH_MAX)
+        if dscp is not None and not 0 <= dscp <= const.QOS_DSCP_MAX:
+            raise ValueError("dscp(%s) must be between 0 and %s, inclusive" %
+                             dscp, const.QOS_DSCP_MAX)
+        if rate is None and dscp is None:
+            raise ValueError("One of the rate or dscp must be configured")
+        super(QoSAddCommand, self).__init__(api)
+        self.switch = switch
+        self.direction = direction
+        self.priority = priority
+        self.match = match
+        self.rate = rate
+        self.burst = burst
+        self.dscp = dscp
+        self.may_exist = may_exist
+        self.columns = columns
+
+    def qos_match(self, row):
+        return (self.direction == row.direction and
+                self.priority == row.priority and
+                self.match == row.match)
+
+    def run_idl(self, txn):
+        ls = self.api.lookup('Logical_Switch', self.switch)
+        qos_rules = [row for row in ls.qos_rules if self.qos_match(row)]
+        if qos_rules:
+            if self.may_exist:
+                self.result = rowview.RowView(qos_rules[0])
+                return
+            raise RuntimeError("QoS (%s, %s, %s) already exists" % (
+                self.direction, self.priority, self.match))
+        row = txn.insert(self.api.tables[self.table_name])
+        row.direction = self.direction
+        row.priority = self.priority
+        row.match = self.match
+        if self.rate:
+            row.setkey('bandwidth', 'rate', self.rate)
+            if self.burst:
+                row.setkey('bandwidth', 'burst', self.burst)
+        if self.dscp is not None:
+            row.setkey('action', 'dscp', self.dscp)
+        self.set_columns(row, **self.columns)
+        ls.addvalue('qos_rules', row)
+        self.result = row.uuid
+
+
+class QoSDelCommand(cmd.BaseCommand):
     def __init__(self, api, switch, direction=None,
                  priority=None, match=None):
         if (priority is None) != (match is None):
             raise TypeError("Must specify priority and match together")
         if priority is not None and not direction:
             raise TypeError("Cannot specify priority/match without direction")
-        super(AclDelCommand, self).__init__(api)
+        super(QoSDelCommand, self).__init__(api)
         self.switch = switch
         self.conditions = []
         if direction:
             self.conditions.append(('direction', '=', direction))
             # priority can be 0
-            if match:  # and therefor prioroity due to the above check
+            if match:  # and therefor priority due to the above check
                 self.conditions += [('priority', '=', priority),
                                     ('match', '=', match)]
 
     def run_idl(self, txn):
         ls = self.api.lookup('Logical_Switch', self.switch)
-        for acl in [a for a in ls.acls
-                    if idlutils.row_match(a, self.conditions)]:
-            ls.delvalue('acls', acl)
-            acl.delete()
+        for row in ls.qos_rules:
+            if idlutils.row_match(row, self.conditions):
+                ls.delvalue('qos_rules', row)
+                row.delete()
 
 
-class AclListCommand(cmd.BaseCommand):
+class QoSListCommand(cmd.BaseCommand):
     def __init__(self, api, switch):
-        super(AclListCommand, self).__init__(api)
+        super(QoSListCommand, self).__init__(api)
         self.switch = switch
 
     def run_idl(self, txn):
         ls = self.api.lookup('Logical_Switch', self.switch)
-        self.result = [rowview.RowView(acl) for acl in ls.acls]
+        self.result = [rowview.RowView(row) for row in ls.qos_rules]
 
 
 class LspAddCommand(cmd.AddCommand):
@@ -296,13 +438,14 @@ class LspGetTagCommand(cmd.BaseCommand):
 
 class LspSetAddressesCommand(cmd.BaseCommand):
     addr_re = re.compile(
-        r'^(router|unknown|dynamic|([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2} .+)$')
+        r'^(router|unknown|dynamic|([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}( .+)*)$')
 
     def __init__(self, api, port, addresses):
         for addr in addresses:
             if not self.addr_re.match(addr):
                 raise TypeError(
-                    "address must be router/unknown/dynamic/ethaddr ipaddr...")
+                    "address must be router/unknown/dynamic/"
+                    "ethaddr[ ipaddr...]")
         super(LspSetAddressesCommand, self).__init__(api)
         self.port = port
         self.addresses = addresses
@@ -550,6 +693,10 @@ class LrListCommand(cmd.BaseCommand):
                        r in self.api.tables['Logical_Router'].rows.values()]
 
 
+class LrGetCommand(cmd.BaseGetRowCommand):
+    table = 'Logical_Router'
+
+
 class LrpAddCommand(cmd.BaseCommand):
     def __init__(self, api, router, port, mac, networks,
                  peer=None, may_exist=False, **columns):
@@ -1051,3 +1198,97 @@ class DnsSetExternalIdsCommand(cmd.BaseCommand):
         except idlutils.RowNotFound:
             msg = "DNS %s does not exist" % self.row_uuid
             raise RuntimeError(msg)
+
+
+class PgAddCommand(cmd.AddCommand):
+    table_name = 'Port_Group'
+
+    def __init__(self, api, name, may_exist=False, **columns):
+        super(PgAddCommand, self).__init__(api)
+        self.name = name
+        self.may_exist = may_exist
+        self.columns = columns
+
+    def run_idl(self, txn):
+        if self.may_exist:
+            try:
+                pg = self.api.lookup(self.table_name, self.name)
+                self.result = rowview.RowView(pg)
+                return
+            except idlutils.RowNotFound:
+                pass
+
+        pg = txn.insert(self.api._tables[self.table_name])
+        pg.name = self.name or ""
+        self.set_columns(pg, **self.columns)
+        self.result = pg.uuid
+
+
+class PgDelCommand(cmd.BaseCommand):
+    table_name = 'Port_Group'
+
+    def __init__(self, api, name, if_exists=False):
+        super(PgDelCommand, self).__init__(api)
+        self.name = name
+        self.if_exists = if_exists
+
+    def run_idl(self, txn):
+        try:
+            pg = self.api.lookup(self.table_name, self.name)
+            pg.delete()
+        except idlutils.RowNotFound:
+            if self.if_exists:
+                return
+            raise RuntimeError('Port group %s does not exist' % self.name)
+
+
+class _PgUpdatePortsHelper(cmd.BaseCommand):
+    method = None
+
+    def __init__(self, api, port_group, lsp=None, if_exists=False):
+        super(_PgUpdatePortsHelper, self).__init__(api)
+        self.port_group = port_group
+        self.lsp = [] if lsp is None else self._listify(lsp)
+        self.if_exists = if_exists
+
+    def _listify(self, res):
+        return res if isinstance(res, (list, tuple)) else [res]
+
+    def _run_method(self, pg, port):
+        if not port:
+            return
+
+        if isinstance(port, cmd.BaseCommand):
+            port = port.result
+        elif uuidutils.is_uuid_like(port):
+            try:
+                port = self.api.lookup('Logical_Switch_Port', port)
+            except idlutils.RowNotFound:
+                if self.if_exists:
+                    return
+                raise RuntimeError(
+                    'Port %s does not exist' % port)
+
+        getattr(pg, self.method)('ports', port)
+
+    def run_idl(self, txn):
+        try:
+            pg = self.api.lookup('Port_Group', self.port_group)
+        except idlutils.RowNotFound:
+            raise RuntimeError('Port group %s does not exist' %
+                               self.port_group)
+
+        for lsp in self.lsp:
+            self._run_method(pg, lsp)
+
+
+class PgAddPortCommand(_PgUpdatePortsHelper):
+    method = 'addvalue'
+
+
+class PgDelPortCommand(_PgUpdatePortsHelper):
+    method = 'delvalue'
+
+
+class PgGetCommand(cmd.BaseGetRowCommand):
+    table = 'Port_Group'
