@@ -11,6 +11,7 @@
 #    under the License.
 
 import netaddr
+from oslo_utils import uuidutils
 import testscenarios
 
 from ovsdbapp.backend.ovs_idl import idlutils
@@ -107,11 +108,19 @@ class TestAclOps(OvnNorthboundTest):
     def setUp(self):
         super(TestAclOps, self).setUp()
         self.switch = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        self.port_group = self.useFixture(fixtures.PortGroupFixture()).obj
+
+    def _acl_add(self, entity, *args, **kwargs):
+        self.assertIn(entity, ['lswitch', 'port_group'])
+        if entity == 'lswitch':
+            cmd = self.api.acl_add(self.switch.uuid, *args, **kwargs)
+            resource = self.switch
+        else:
+            cmd = self.api.pg_acl_add(self.port_group.uuid, *args, **kwargs)
+            resource = self.port_group
 
-    def _acl_add(self, *args, **kwargs):
-        cmd = self.api.acl_add(self.switch.uuid, *args, **kwargs)
         aclrow = cmd.execute(check_error=True)
-        self.assertIn(aclrow._row, self.switch.acls)
+        self.assertIn(aclrow._row, resource.acls)
         self.assertEqual(cmd.direction, aclrow.direction)
         self.assertEqual(cmd.priority, aclrow.priority)
         self.assertEqual(cmd.match, aclrow.match)
@@ -119,43 +128,50 @@ class TestAclOps(OvnNorthboundTest):
         return aclrow
 
     def test_acl_add(self):
-        self._acl_add('from-lport', 0, 'output == "fake_port" && ip',
-                      'drop')
+        self._acl_add('lswitch', 'from-lport', 0,
+                      'output == "fake_port" && ip', 'drop')
 
     def test_acl_add_exists(self):
-        args = ('from-lport', 0, 'output == "fake_port" && ip', 'drop')
+        args = ('lswitch', 'from-lport', 0, 'output == "fake_port" && ip',
+                'drop')
         self._acl_add(*args)
         self.assertRaises(RuntimeError, self._acl_add, *args)
 
     def test_acl_add_may_exist(self):
         args = ('from-lport', 0, 'output == "fake_port" && ip', 'drop')
-        row = self._acl_add(*args)
-        row2 = self._acl_add(*args, may_exist=True)
+        row = self._acl_add('lswitch', *args)
+        row2 = self._acl_add('lswitch', *args, may_exist=True)
         self.assertEqual(row, row2)
 
     def test_acl_add_extids(self):
         external_ids = {'mykey': 'myvalue', 'yourkey': 'yourvalue'}
-        acl = self._acl_add('from-lport', 0, 'output == "fake_port" && ip',
+        acl = self._acl_add('lswitch',
+                            'from-lport', 0, 'output == "fake_port" && ip',
                             'drop', **external_ids)
         self.assertEqual(external_ids, acl.external_ids)
 
     def test_acl_del_all(self):
-        r1 = self._acl_add('from-lport', 0, 'output == "fake_port"', 'drop')
+        r1 = self._acl_add('lswitch', 'from-lport', 0, 'output == "fake_port"',
+                           'drop')
         self.api.acl_del(self.switch.uuid).execute(check_error=True)
         self.assertNotIn(r1.uuid, self.api.tables['ACL'].rows)
         self.assertEqual([], self.switch.acls)
 
     def test_acl_del_direction(self):
-        r1 = self._acl_add('from-lport', 0, 'output == "fake_port"', 'drop')
-        r2 = self._acl_add('to-lport', 0, 'output == "fake_port"', 'allow')
+        r1 = self._acl_add('lswitch', 'from-lport', 0,
+                           'output == "fake_port"', 'drop')
+        r2 = self._acl_add('lswitch', 'to-lport', 0,
+                           'output == "fake_port"', 'allow')
         self.api.acl_del(self.switch.uuid, 'from-lport').execute(
             check_error=True)
         self.assertNotIn(r1, self.switch.acls)
         self.assertIn(r2, self.switch.acls)
 
     def test_acl_del_direction_priority_match(self):
-        r1 = self._acl_add('from-lport', 0, 'output == "fake_port"', 'drop')
-        r2 = self._acl_add('from-lport', 1, 'output == "fake_port"', 'allow')
+        r1 = self._acl_add('lswitch', 'from-lport', 0,
+                           'output == "fake_port"', 'drop')
+        r2 = self._acl_add('lswitch', 'from-lport', 1,
+                           'output == "fake_port"', 'allow')
         cmd = self.api.acl_del(self.switch.uuid,
                                'from-lport', 0, 'output == "fake_port"')
         cmd.execute(check_error=True)
@@ -171,12 +187,135 @@ class TestAclOps(OvnNorthboundTest):
                           priority=0)
 
     def test_acl_list(self):
-        r1 = self._acl_add('from-lport', 0, 'output == "fake_port"', 'drop')
-        r2 = self._acl_add('from-lport', 1, 'output == "fake_port2"', 'allow')
+        r1 = self._acl_add('lswitch', 'from-lport', 0,
+                           'output == "fake_port"', 'drop')
+        r2 = self._acl_add('lswitch', 'from-lport', 1,
+                           'output == "fake_port2"', 'allow')
         acls = self.api.acl_list(self.switch.uuid).execute(check_error=True)
         self.assertIn(r1, acls)
         self.assertIn(r2, acls)
 
+    def test_pg_acl_add(self):
+        self._acl_add('port_group', 'from-lport', 0,
+                      'output == "fake_port" && ip', 'drop')
+
+    def test_pg_acl_del_all(self):
+        r1 = self._acl_add('port_group', 'from-lport', 0,
+                           'output == "fake_port"', 'drop')
+        self.api.pg_acl_del(self.port_group.uuid).execute(check_error=True)
+        self.assertNotIn(r1.uuid, self.api.tables['ACL'].rows)
+        self.assertEqual([], self.port_group.acls)
+
+    def test_pg_acl_list(self):
+        r1 = self._acl_add('port_group', 'from-lport', 0,
+                           'output == "fake_port"', 'drop')
+        r2 = self._acl_add('port_group', 'from-lport', 1,
+                           'output == "fake_port2"', 'allow')
+        acls = self.api.pg_acl_list(self.port_group.uuid).execute(
+            check_error=True)
+        self.assertIn(r1, acls)
+        self.assertIn(r2, acls)
+
+
+class TestQoSOps(OvnNorthboundTest):
+    def setUp(self):
+        super(TestQoSOps, self).setUp()
+        self.switch = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+
+    def _qos_add(self, *args, **kwargs):
+        cmd = self.api.qos_add(self.switch.uuid, *args, **kwargs)
+        row = cmd.execute(check_error=True)
+        self.assertIn(row._row, self.switch.qos_rules)
+        self.assertEqual(cmd.direction, row.direction)
+        self.assertEqual(cmd.priority, row.priority)
+        self.assertEqual(cmd.match, row.match)
+        self.assertEqual(cmd.rate, row.bandwidth.get('rate', None))
+        self.assertEqual(cmd.burst, row.bandwidth.get('burst', None))
+        self.assertEqual(cmd.dscp, row.action.get('dscp', None))
+        return row
+
+    def test_qos_add_dscp(self):
+        self._qos_add('from-lport', 0, 'output == "fake_port" && ip', dscp=33)
+
+    def test_qos_add_rate(self):
+        self._qos_add('from-lport', 0, 'output == "fake_port" && ip', rate=100)
+
+    def test_qos_add_rate_burst(self):
+        self._qos_add('from-lport', 0, 'output == "fake_port" && ip', rate=101,
+                      burst=1001)
+
+    def test_qos_add_rate_dscp(self):
+        self._qos_add('from-lport', 0, 'output == "fake_port" && ip', rate=102,
+                      burst=1002, dscp=56)
+
+    def test_qos_add_raises(self):
+        self.assertRaises(TypeError, self.api.qos_add, 'from-lport', 0,
+                          'output == "fake_port" && ip')
+
+    def test_qos_add_direction_raises(self):
+        self.assertRaises(TypeError, self.api.qos_add, 'foo', 0, 'ip',
+                          bandwidth={'rate': 102, 'burst': 1002})
+
+    def test_qos_add_priority_raises(self):
+        self.assertRaises(TypeError, self.api.qos_add, 'from-lport', 32768,
+                          'ip', bandwidth={'rate': 102, 'burst': 1002})
+
+    def test_qos_add_exists(self):
+        args = ('from-lport', 0, 'output == "fake_port" && ip', 1000)
+        self._qos_add(*args)
+        self.assertRaises(RuntimeError, self._qos_add, *args)
+
+    def test_qos_add_may_exist(self):
+        args = ('from-lport', 0, 'output == "fake_port" && ip', 1000)
+        row = self._qos_add(*args)
+        row2 = self._qos_add(*args, may_exist=True)
+        self.assertEqual(row, row2)
+
+    def test_qos_add_extids(self):
+        external_ids = {'mykey': 'myvalue', 'yourkey': 'yourvalue'}
+        qos = self._qos_add('from-lport', 0, 'output == "fake_port" && ip',
+                            dscp=11, external_ids=external_ids)
+        self.assertEqual(external_ids, qos.external_ids)
+
+    def test_qos_del_all(self):
+        r1 = self._qos_add('from-lport', 0, 'output == "fake_port"', 1000)
+        self.api.qos_del(self.switch.uuid).execute(check_error=True)
+        self.assertNotIn(r1.uuid, self.api.tables['QoS'].rows)
+        self.assertEqual([], self.switch.qos_rules)
+
+    def test_qos_del_direction(self):
+        r1 = self._qos_add('from-lport', 0, 'output == "fake_port"', 1000)
+        r2 = self._qos_add('to-lport', 0, 'output == "fake_port"', 1000)
+        self.api.qos_del(self.switch.uuid, 'from-lport').execute(
+            check_error=True)
+        self.assertNotIn(r1, self.switch.qos_rules)
+        self.assertIn(r2, self.switch.qos_rules)
+
+    def test_qos_del_direction_priority_match(self):
+        r1 = self._qos_add('from-lport', 0, 'output == "fake_port"', 1000)
+        r2 = self._qos_add('from-lport', 1, 'output == "fake_port"', 1000)
+        cmd = self.api.qos_del(self.switch.uuid,
+                               'from-lport', 0, 'output == "fake_port"')
+        cmd.execute(check_error=True)
+        self.assertNotIn(r1, self.switch.qos_rules)
+        self.assertIn(r2, self.switch.qos_rules)
+
+    def test_qos_del_priority_without_match(self):
+        self.assertRaises(TypeError, self.api.qos_del, self.switch.uuid,
+                          'from-lport', 0)
+
+    def test_qos_del_priority_without_direction(self):
+        self.assertRaises(TypeError, self.api.qos_del, self.switch.uuid,
+                          priority=0)
+
+    def test_qos_list(self):
+        r1 = self._qos_add('from-lport', 0, 'output == "fake_port"', 1000)
+        r2 = self._qos_add('from-lport', 1, 'output == "fake_port2"', 1000)
+        qos_rules = self.api.qos_list(self.switch.uuid).execute(
+            check_error=True)
+        self.assertIn(r1, qos_rules)
+        self.assertIn(r2, qos_rules)
+
 
 class TestLspOps(OvnNorthboundTest):
     def setUp(self):
@@ -297,8 +436,8 @@ class TestLspOps(OvnNorthboundTest):
 
     def test_lsp_del_wrong_switch(self):
         lsp = self._lsp_add(self.switch, None)
-        sw_id = self.useFixture(fixtures.LogicalSwitchFixture()).obj
-        cmd = self.api.lsp_del(lsp.uuid, sw_id)
+        sw = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        cmd = self.api.lsp_del(lsp.uuid, sw.uuid)
         self.assertRaises(RuntimeError, cmd.execute, check_error=True)
 
     def test_lsp_del_switch_no_exist(self):
@@ -342,7 +481,7 @@ class TestLspOps(OvnNorthboundTest):
 
     def test_lsp_set_addresses(self):
         lsp = self._lsp_add(self.switch, None)
-        for addr in ('dynamic', 'unknown', 'router',
+        for addr in ('dynamic', 'unknown', 'router', 'de:ad:be:ef:4d:ad',
                      'de:ad:be:ef:4d:ad 192.0.2.1'):
             self.api.lsp_set_addresses(lsp.name, [addr]).execute(
                 check_error=True)
@@ -351,7 +490,7 @@ class TestLspOps(OvnNorthboundTest):
     def test_lsp_set_addresses_invalid(self):
         self.assertRaises(
             TypeError,
-            self.api.lsp_set_addresses, 'fake', '01:02:03:04:05:06')
+            self.api.lsp_set_addresses, 'fake', 'invalidaddress')
 
     def test_lsp_get_addresses(self):
         addresses = [
@@ -825,8 +964,8 @@ class TestLogicalRouterPortOps(OvnNorthboundTest):
 
     def test_lrp_del_wrong_router(self):
         lrp = self._lrp_add(None)
-        sw_id = self.useFixture(fixtures.LogicalSwitchFixture()).obj
-        cmd = self.api.lrp_del(lrp.uuid, sw_id)
+        sw = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        cmd = self.api.lrp_del(lrp.uuid, sw.uuid)
         self.assertRaises(RuntimeError, cmd.execute, check_error=True)
 
     def test_lrp_del_router_no_exist(self):
@@ -1197,3 +1336,73 @@ class TestLsDnsOps(OvnNorthboundTest):
                               [dns.uuid for dns in ls1.dns_records])
         self.api.ls_remove_dns_record(ls1.uuid, dns1.uuid).execute()
         self.assertEqual([], ls1.dns_records)
+
+
+class TestPortGroup(OvnNorthboundTest):
+
+    def setUp(self):
+        super(TestPortGroup, self).setUp()
+        self.switch = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        self.pg_name = 'testpg-%s' % uuidutils.generate_uuid()
+
+    def test_port_group(self):
+        # Assert the Port Group was added
+        self.api.pg_add(self.pg_name).execute(check_error=True)
+        row = self.api.db_find(
+            'Port_Group',
+            ('name', '=', self.pg_name)).execute(check_error=True)
+        self.assertIsNotNone(row)
+        self.assertEqual(self.pg_name, row[0]['name'])
+        self.assertEqual([], row[0]['ports'])
+        self.assertEqual([], row[0]['acls'])
+
+        # Assert the Port Group was deleted
+        self.api.pg_del(self.pg_name).execute(check_error=True)
+        row = self.api.db_find(
+            'Port_Group',
+            ('name', '=', self.pg_name)).execute(check_error=True)
+        self.assertEqual([], row)
+
+    def test_port_group_ports(self):
+        lsp_add_cmd = self.api.lsp_add(self.switch.uuid, 'testport')
+        with self.api.transaction(check_error=True) as txn:
+            txn.add(lsp_add_cmd)
+            txn.add(self.api.pg_add(self.pg_name))
+
+        port_uuid = lsp_add_cmd.result.uuid
+
+        # Lets add the port using the UUID instead of a `Command` to
+        # exercise the API
+        self.api.pg_add_ports(self.pg_name, port_uuid).execute(
+            check_error=True)
+        row = self.api.db_find(
+            'Port_Group',
+            ('name', '=', self.pg_name)).execute(check_error=True)
+        self.assertIsNotNone(row)
+        self.assertEqual(self.pg_name, row[0]['name'])
+        # Assert the port was added from the Port Group
+        self.assertEqual([port_uuid], row[0]['ports'])
+
+        # Delete the Port from the Port Group
+        with self.api.transaction(check_error=True) as txn:
+            txn.add(self.api.pg_del_ports(self.pg_name, port_uuid))
+
+        row = self.api.db_find(
+            'Port_Group',
+            ('name', '=', self.pg_name)).execute(check_error=True)
+        self.assertIsNotNone(row)
+        self.assertEqual(self.pg_name, row[0]['name'])
+        # Assert the port was removed from the Port Group
+        self.assertEqual([], row[0]['ports'])
+
+    def test_pg_del_ports_if_exists(self):
+        self.api.pg_add(self.pg_name).execute(check_error=True)
+        non_existent_res = uuidutils.generate_uuid()
+
+        # Assert that if if_exists is False (default) it will raise an error
+        self.assertRaises(RuntimeError, self.api.pg_del_ports(self.pg_name,
+                          non_existent_res).execute, True)
+
+        # Assert that if if_exists is True it won't raise an error
+        self.api.pg_del_ports(self.pg_name, non_existent_res,
+                              if_exists=True).execute(check_error=True)
